// Copyright (C) 2008-2012, Software Development Laboratories, "Fish" (David B. Trout)
//////////////////////////////////////////////////////////////////////////////////////////
// ParseOpt.cpp   --   custom command-line parser
//////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  04/19/08    1.0.0   Fish    Created.
//  03/17/09    1.1.0   Fish    -rcfile support.
//  08/20/09    1.1.2   Fish    Banner tweak.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "ParseOpt.h"

//////////////////////////////////////////////////////////////////////////////////////////

void CMyCmdLineParser::ShowHelp( FILE* pStream /* =stdout */ )
{
    LPCTSTR  pszHelp
        =
        _T("\n")
        _T("                      ") _T(PROGNAME) _T("\n")
        _T("          ") _T(PRODUCT) _T("\n")
        _T("                 Version ") VERSION_STR _T("\n")
        _T("\n")
        _T("  ") _T(COPYRIGHT) _T(", ") _T(COMPANY) _T("\n")
        _T("         ") _T(PRODUCT_URL) _T("\n")
        _T("                ") _T(PRODUCT_EMAIL) _T("\n")
        _T("\n")
        _T("Format:\n")
        _T("\n")
        _T("   ") _T(PROGNAME) _T("   <jobsfile>   [ -j nnn ]   [ -rcfile xxxxx ]\n")
        _T("\n")
        _T("\n")
        _T("Where:\n")
        _T("\n")
        _T("   <jobsfile>      filename of command-lines to be started\n")
        _T("   -j nnn          maximum number of simultaneous commands\n")
        _T("   -rcfile xxxxx   name of optional return codes file\n")
        _T("\n")
        _T("\n")
        _T("The jobs file is a simple verbatim list of the command-lines\n")
        _T("that you wish to issue.\n")
        _T("\n")
        _T("Lines whose first non-blank character is '#', ';' or '*' are\n")
        _T("considered comment lines and are ignored. All blank lines are\n")
        _T("also ignored.\n")
        _T("\n")
        _T("All other lines contain the exact command-line to be issued.\n")
        _T("\n")
        _T("The -j parameter sets an upper limit to the maximum number of\n")
        _T("simultaneously running commands.\n")
        _T("\n")
        _T("When the jobs file contains more than this number of commands\n")
        _T("the next job (command) is not started until one or more of the\n")
        _T("previously issued jobs (commands) finishes first.\n")
        _T("\n")
        _T("Specifying 0 (zero) indicates no upper limit (i.e. ALL listed\n")
        _T("jobs will be started simultaneously regardless of how ever many\n")
        _T("there are).\n")
        _T("\n")
        _T("The default for -j is the value of the \"RUNJOBS_DEFAULT_MAXJOBS\"\n")
        _T("environment variable if defined or the number of CPUs installed\n")
        _T("on your system if it is not defined.\n")
        _T("\n")
        _T("The return code from ") _T(PROGNAME) _T(" is always the maximum returned\n")
        _T("return code value from any of the started jobs. If the -rcfile\n")
        _T("option is specified the return code value from each job is also\n")
        _T("written to the specified return codes file using the following\n")
        _T("format: \"n = rc\" where 'n' is the job number and 'rc' is its\n")
        _T("corresponding return code value.\n")
        _T("\n")
        _T("\n")
        _T("                        USAGE NOTE\n")
        _T("\n")
        _T("Since RunJobs doesn't parse jobs file command-lines, all of your\n")
        _T("environment variable strings on your command-line SHOULD have a\n")
        _T("valid (non-empty) value assigned to it. Otherwise an improperly\n")
        _T("formed command-line will be built since ExpandEnvironmentStrings\n")
        _T("leaves the variable unexpanded whenever it cannot find it. For\n")
        _T("example: if your specified command-line is:\n")
        _T("\n")
        _T("    VCBUILD  %VCBUILDTYPE%  \"MyProj.vcproj\"  \"Debug|Win32\"\n")
        _T("\n")
        _T("then VCBUILD fails (and crashes if invoked within the IDE!) with\n")
        _T("the following error:\n")
        _T("\n")
        _T("  \"vcbuild.exe : warning VCBLD6002: invalid option Debug|Win32\n")
        _T("                 specified. The option was ignored.\"\n")
        _T("\n")
        _T("To prevent this, ensure that all environment variable strings on\n")
        _T("your command-line have a value assigned to them.\n")
        _T("\n")
        ;

    WriteToStdXxx( pStream, 0, pszHelp );
}

//////////////////////////////////////////////////////////////////////////////////////////

void CMyCmdLineParser::ParseParam( LPCTSTR pszParam, BOOL bFlag, BOOL bLast )
{
    static bool bMaxJobs = false, bRCFile = false;
    CString strParam = pszParam;
    CString strErrMsg;

    if (bMaxJobs || bRCFile)
    {
        ASSERT(!(bMaxJobs && bRCFile)); // (should never have both on at same time!)

        // The previous parameter was a flag indicating the next parameter
        // was going to be either the maxjobs value or the rcfile value...

        if (bFlag)
        {
            // The "next parameter" (i.e. THIS parameter, i.e. the CURRENT parameter)
            // is ALSO a flag, meaning the maxjobs value or the rcfile value that we
            // THOUGHT was going to be specified next was in fact NOT specified.

            if (bMaxJobs)
                strErrMsg = _T(PROGNAME) _T(": ERROR: missing maxjobs value.\n");
            else // (bRCFile)
                strErrMsg = _T(PROGNAME) _T(": ERROR: missing rcfiles value.\n");

            TRACE( _T("%s"), (LPCTSTR) strErrMsg );
            WriteToStdXxx( stderr, 0, strErrMsg );
            m_bErrors = true;

            bMaxJobs = false;
            bRCFile  = false;

            // Now parse the current flag...

            ParseParam( pszParam, bFlag, bLast );
        }
        else
        {
            // The current parameter is either the expected maxjobs value
            // or else is the expected rcfiles value...

            if (bMaxJobs)
            {
                int nMaxJobs;

                if (1
                    && strParam.SpanIncluding( _T("0123456789") ) == strParam
                    && (nMaxJobs = _ttoi( strParam )) >= 0
                    && ERANGE != errno
                )
                {
                    g_dwMaxJobs = nMaxJobs;
                }
                else
                {
                    strErrMsg.Format
                    (
                        _T(PROGNAME) _T(": ERROR: invalid maxjobs value \"%s\".\n")
                        ,strParam
                    );
                    TRACE( _T("%s"), (LPCTSTR) strErrMsg );
                    WriteToStdXxx( stderr, 0, strErrMsg );
                    m_bErrors = true;
                }
            }
            else // (bRCFile)
            {
                g_strRCFileFilename = strParam;
            }

            bMaxJobs = false;
            bRCFile  = false;
        }
        return;
    }

    ASSERT(!bMaxJobs && !bRCFile);  // (both should have been handled just above!)

    // Flag?

    if (bFlag)
    {
        if (strParam.CompareNoCase(_T("?")) == 0)
        {
            m_bErrors = true; // (force mainline to call ShowHelp)
        }
        else if (strParam.CompareNoCase(_T("j")) == 0)
        {
            // Maxjobs option...

            if (!bLast && !bRCFile)
            {
                bMaxJobs = true;
            }
            else // (bLast || bRCFile)
            {
                // The current -j flag is unexpected. We were either expecting the
                // rcfile value to be specified next (because our bRCFile flag is on)
                // or else the "-j" flag is the very last string on the command-line
                // meaning the required maxjobs value is missing, or both...

                if (bRCFile)
                {
                    strErrMsg = _T(PROGNAME) _T(": ERROR: missing rcfiles value.\n");
                    TRACE( _T("%s"), (LPCTSTR) strErrMsg );
                    WriteToStdXxx( stderr, 0, strErrMsg );
                }

                if (bLast)
                {
                    strErrMsg = _T(PROGNAME) _T(": ERROR: missing maxjobs value.\n");
                    TRACE( _T("%s"), (LPCTSTR) strErrMsg );
                    WriteToStdXxx( stderr, 0, strErrMsg );
                }

                m_bErrors = true;
            }
        }
        else if (strParam.CompareNoCase(_T("rcfile")) == 0)
        {
            // RCFile option...

            if (!bLast && !bMaxJobs)
            {
                bRCFile = true;
            }
            else // (bLast || bMaxJobs)
            {
                // The current -rcfile option is unexpected. We were either expecting
                // the maxjobs value to be specified next (because our bMaxJobs flag
                // is on) or else the "-rcfile" option was specified as the very last
                // string on the command-line meaning the required filename is missing,
                // OR *both* conditions are true...

                if (bMaxJobs)
                {
                    strErrMsg = _T(PROGNAME) _T(": ERROR: missing maxjobs value.\n");
                    TRACE( _T("%s"), (LPCTSTR) strErrMsg );
                    WriteToStdXxx( stderr, 0, strErrMsg );
                }

                if (bLast)
                {
                    strErrMsg = _T(PROGNAME) _T(": ERROR: missing rcfiles value.\n");
                    TRACE( _T("%s"), (LPCTSTR) strErrMsg );
                    WriteToStdXxx( stderr, 0, strErrMsg );
                }

                m_bErrors = true;
            }
        }
        else
        {
            // It's not an option (flag) we support...

            strErrMsg.Format
            (
                _T(PROGNAME) _T(": ERROR: unsupported/unrecognized option \"-%s\".\n")
                ,strParam
            );
        }
        return;
    }

    // It's a standalone (non-flag) parameter...

    if (g_strRunJobsFilename.IsEmpty())
    {
        g_strRunJobsFilename = strParam;
        return;
    }

    strErrMsg.Format
    (
        _T(PROGNAME) _T(": ERROR: extraneous parameter \"%s\".\n")
        ,strParam
    );

    TRACE( _T("%s"), (LPCTSTR) strErrMsg );
    WriteToStdXxx( stderr, 0, strErrMsg );
    m_bErrors = true;
}

//////////////////////////////////////////////////////////////////////////////////////////
