// Copyright (C) 2008-2012, Software Development Laboratories, "Fish" (David B. Trout)
/////////////////////////////////////////////////////////////////////////////////////////
// PipedProcessCtl.cpp  --  handy class to create process with redirected stdxxx pipes
/////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  04/19/08    1.0.0   Fish    Created.
//  11/06/08    1.0.2   Fish    Use UINT instead of int.
//  03/17/09    1.0.3   Fish    Bug fix: Save child's m_dwProcessId!
//  03/17/09    1.1.0   Fish    Use 0 (zero) for flags argument in CreateProcess call
//                              instead of CREATE_NO_WINDOW so cancel gets propagated.
//  08/11/09    1.1.1   Fish    CreateProcess failure message tweak.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "PipedProcessCtl.h"

/////////////////////////////////////////////////////////////////////////////////////////
// Constructor

CPipedProcessCtl::CPipedProcessCtl
(
    UINT            nChildId,               // child id#
    LPCTSTR         pszCommandLine,         // command line (with arguments)
    LPCTSTR         pszWorkingDirectory,    // desired working directory
    WRITETOSTDXXX*  pfnWriteToStdXXX        // stream o/p function ptr
)
    // (use the values passed to us)

    : m_nChildId            ( nChildId            )
    , m_strCommandLine      ( pszCommandLine      )
    , m_strWorkingDirectory ( pszWorkingDirectory )
    , m_pfnWriteToStdXXX    ( pfnWriteToStdXXX    )
{
    // Initialize class variables...

    m_hOurReadFromStderr = NULL;
    m_hOurReadFromStdout = NULL;
    m_hChildProcess      = NULL;

    m_pStderrThread      = NULL;
    m_pStdoutThread      = NULL;
    m_pChildEndedThread  = NULL;

    m_dwProcessId        = 0;
    m_dwExitCode         = 0;
    m_bProcessHasExited  = true;

    ASSERT( m_pfnWriteToStdXXX );   // (MUST NOT BE NULL!)
}

/////////////////////////////////////////////////////////////////////////////////////////
// Destructor

CPipedProcessCtl::~CPipedProcessCtl()
{
    // Wait for child to end

    WaitForProcessToExit();     // (it will close the handles for us)

    // Clean up after ourselves...

    delete  m_pStderrThread;
    delete  m_pStdoutThread;
    delete  m_pChildEndedThread;
}

/////////////////////////////////////////////////////////////////////////////////////////
// Spin off a child process and associated pipe-monitoring worker threads...

bool CPipedProcessCtl::StartChildProcess()      // (WARNING: NOT REUSABLE!)
{
    /////////////////////////////////////////////////////////////////////////////
    //                      Create the pipes...
    /////////////////////////////////////////////////////////////////////////////

    HANDLE hOurProcess = GetCurrentProcess();   // (temporary for creating pipes)
    HANDLE hPipeReadHandle, hPipeWriteHandle;   // (temporary for creating pipes)

    HANDLE hChildReadFromStdin;     // child's stdin  handle (inherited from us)
    HANDLE hChildWriteToStdout;     // child's stdout handle (inherited from us)
    HANDLE hChildWriteToStderr;     // child's stderr handle (inherited from us)

    HANDLE hOurWriteToStdin;        // (temp; will be closed as we don't use it)

    SECURITY_ATTRIBUTES  saAttr;    // (so we can allow handles to be inherited)

    saAttr.nLength               = NUM_BYTES( SECURITY_ATTRIBUTES );
    saAttr.bInheritHandle        = TRUE; // (allows handles to be inherited by child)
    saAttr.lpSecurityDescriptor  = NULL; // (default security)

    // Create a feed pipe to child's stdin input

    VERIFY( CreatePipe( &hChildReadFromStdin, &hPipeWriteHandle, &saAttr, PIPEBUFSIZE ) );

    VERIFY( DuplicateHandle                     // (create non-inheritable copy for ourselves)
    (
        hOurProcess, hPipeWriteHandle,          // (input handle to be duplicated)
        hOurProcess, &hOurWriteToStdin,         // (output duplicated handle)
        0,                                      // (ignored due to DUPLICATE_SAME_ACCESS)
        FALSE,                                  // (prevents child from inheriting ours)
        DUPLICATE_SAME_ACCESS                   // (same access as original handle)
    ));

    CloseHandle( hPipeWriteHandle );            // (no longer needed; close to prevent child from hanging)

    // Pipe child's stdout output back to us...  (uses same technique as just above)

    VERIFY( CreatePipe( &hPipeReadHandle, &hChildWriteToStdout, &saAttr, PIPEBUFSIZE ) );

    VERIFY( DuplicateHandle
    (
        hOurProcess,   hPipeReadHandle,
        hOurProcess, &m_hOurReadFromStdout,
        0,
        FALSE,
        DUPLICATE_SAME_ACCESS
    ));

    CloseHandle( hPipeReadHandle );

    // Pipe child's stderr output back to us...  (uses same technique as just above)

    VERIFY( CreatePipe( &hPipeReadHandle, &hChildWriteToStderr, &saAttr, PIPEBUFSIZE ) );

    VERIFY( DuplicateHandle
    (
        hOurProcess,   hPipeReadHandle,
        hOurProcess, &m_hOurReadFromStderr,
        0,
        FALSE,
        DUPLICATE_SAME_ACCESS
    ));

    CloseHandle( hPipeReadHandle );

    /////////////////////////////////////////////////////////////////////////////
    //                    Create the child process...
    /////////////////////////////////////////////////////////////////////////////

    STARTUPINFO          siStartInfo;       // (input  to   CreateProcess API)
    PROCESS_INFORMATION  piProcInfo;        // (output from CreateProcess API)

    memset( &siStartInfo, 0, NUM_BYTES( STARTUPINFO         ));
    memset( &piProcInfo,  0, NUM_BYTES( PROCESS_INFORMATION ));

    siStartInfo.cb         = NUM_BYTES( STARTUPINFO );  // (size of structure)
    siStartInfo.dwFlags    = STARTF_USESTDHANDLES;      // (use redirected std handles)
    siStartInfo.hStdInput  = hChildReadFromStdin;       // (use redirected std handles)
    siStartInfo.hStdOutput = hChildWriteToStdout;       // (use redirected std handles)
    siStartInfo.hStdError  = hChildWriteToStderr;       // (use redirected std handles)

    CString strCommandLine = m_strCommandLine + _T("     ");    // (docs state it needs extra!?)
    LPCTSTR pszCommandLine = strCommandLine;                    // (must be writable memory too!?)

    m_bProcessHasExited = false;        // (since we're starting it up)

    if ( !CreateProcess
    (
        NULL,                           // name of executable module (NULL == use cmdline)
        (LPTSTR) pszCommandLine,        // complete command line with all needed arguments
        NULL,                           // process security attributes (NULL == default)
        NULL,                           // primary thread security attributes (NULL == default)
        TRUE,                           // (required when STARTF_USESTDHANDLES flag is used)
        0,                              // magic creation flag!  >;->
        NULL,                           // ptr to environment (NULL == copy parent's)
        m_strWorkingDirectory,          // working directory
        &siStartInfo,                   // STARTUPINFO pointer
        &piProcInfo                     // receives PROCESS_INFORMATION
    ))
    {
        // Something went wrong...

        m_dwExitCode = GetLastError();

        m_bProcessHasExited = true;     // (since it wasn't even started)

        CString strErrMsg; strErrMsg.Format
        (
            _T("%s: CreateProcess() failed: rc=%u (0x%08.8X): %s.\n")

            ,(LPCTSTR) g_strProgName
            ,m_dwExitCode
            ,m_dwExitCode
            ,(LPCTSTR) FormatLastError( m_dwExitCode )
        );

        TRACE( _T("%s"), strErrMsg );
        m_pfnWriteToStdXXX( stderr, m_nChildId, strErrMsg );

        CloseHandle( hChildReadFromStdin );
        CloseHandle( hChildWriteToStdout );
        CloseHandle( hChildWriteToStderr );

        CloseHandle(   hOurWriteToStdin   );
        CloseHandle( m_hOurReadFromStdout );
        CloseHandle( m_hOurReadFromStderr );

        return false;
    }

    // Save handle to child process and close the ones we don't need...

    m_hChildProcess = piProcInfo.hProcess;      // (save this handle so we can wait on it)
    m_dwProcessId   = piProcInfo.dwProcessId;   // (save id too so we can send it signals)

    CloseHandle( piProcInfo.hThread  );         // (close this one since we don't need it)
    CloseHandle( hOurWriteToStdin    );         // (close this one since we don't need it)

    CloseHandle( hChildReadFromStdin );         // (MUST close OUR copy so child won't hang!)
    CloseHandle( hChildWriteToStdout );         // (MUST close OUR copy so child won't hang!)
    CloseHandle( hChildWriteToStderr );         // (MUST close OUR copy so child won't hang!)

    /////////////////////////////////////////////////////////////////////////////
    //                    Create the worker threads...
    /////////////////////////////////////////////////////////////////////////////

    // Note that since non-MFC threads cannot access MFC objects (due to certain
    // restrictions within MFC itself), we create all the threads as MFC threads
    // so they can access various MFC objects when they need to. This doesn't add
    // any overhead (it doesn't make our threads any less efficient), but it DOES
    // ensure our threads are thread-safe at least from from MFC's point of view.

    m_pChildEndedThread = AfxBeginThread    // (create MFC worker thread)
    (
        ChildEndedWorkerThread,             // (ptr to thread function)
        (void*)this,                        // (ptr to thread arguments)
        THREAD_PRIORITY_TIME_CRITICAL,      // (desired thread priority)
        0,                                  // (initial stack size; 0 = default)
        CREATE_SUSPENDED,                   // (creation options)
        NULL                                // (security attributes; NULL = default)
    );

    m_pStdoutThread = AfxBeginThread        // (see above)
    (
        StdoutWorkerThread,
        (void*)this,
        THREAD_PRIORITY_NORMAL,
        0,
        CREATE_SUSPENDED,
        NULL
    );

    m_pStderrThread = AfxBeginThread        // (see above)
    (
        StderrWorkerThread,
        (void*)this,
        THREAD_PRIORITY_NORMAL,
        0,
        CREATE_SUSPENDED,
        NULL
    );

    // Since we wish to wait on the thread handle (which is a member of the CWinThread
    // object itself), we need to ensure that the MFC CWinThread object itself doesn't
    // get automatically deleted out from under us (which is normal behavior whenever
    // the thread ends: it deletes itself). Thus we need to ask MFC to not automatically
    // do that please, making US responsible for deleting the actual CWinThread object.

    m_pChildEndedThread -> m_bAutoDelete = FALSE;           // (we delete it ourself)
    m_pStderrThread     -> m_bAutoDelete = FALSE;           // (we delete it ourself)
    m_pStdoutThread     -> m_bAutoDelete = FALSE;           // (we delete it ourself)

    VERIFY( m_pChildEndedThread -> ResumeThread() == 1 );   // (now start the thread)
    VERIFY( m_pStdoutThread     -> ResumeThread() == 1 );   // (now start the thread)
    VERIFY( m_pStderrThread     -> ResumeThread() == 1 );   // (now start the thread)

    return true;
}

/////////////////////////////////////////////////////////////////////////////////////////
// Worker thread -- waits forever for child process to end...

UINT ChildEndedWorkerThread( void* pParms )     // (created as MFC worker thread)
{
    CPipedProcessCtl*    pCtl  = (CPipedProcessCtl*) pParms;
    WaitForSingleObject( pCtl->m_hChildProcess, INFINITE );
    GetExitCodeProcess(  pCtl->m_hChildProcess, &pCtl->m_dwExitCode );
                         pCtl->m_bProcessHasExited = true;
    return 0;
}

/////////////////////////////////////////////////////////////////////////////////////////
// Child stdout/stderr worker threads

UINT StdoutWorkerThread( void* pParms )     // (created as MFC worker thread)
{
    return MonitorStdXxxPipe( STD_OUTPUT_HANDLE, (CPipedProcessCtl*) pParms );
}

UINT StderrWorkerThread( void* pParms )     // (created as MFC worker thread)
{
    return MonitorStdXxxPipe( STD_ERROR_HANDLE, (CPipedProcessCtl*) pParms );
}

/////////////////////////////////////////////////////////////////////////////////////////
// (static vars)

static LPCTSTR     g_pszBufferOverflowMsg   = _T("...(truncated)\n");
static const UINT  g_nBufferOverflowMsgLen  = (UINT) _tcslen( g_pszBufferOverflowMsg ); // (#of CHARS, not bytes!)

/////////////////////////////////////////////////////////////////////////////////////////
// (common StdxxxWorkerThread logic)

UINT MonitorStdXxxPipe( DWORD nWhichStdXxx, CPipedProcessCtl* pCtl )
{
    // This function executes in the context of the worker thread that calls it...

    HANDLE  hOurReadFromStdXxx  = NULL;     // (which of the child's streams we're to read from)
    FILE*   pStream             = NULL;     // (which stream (stdout/stderr) we're to write to)
    DWORD   nCharsRead          = 0;        // (#of WHOLE CHARS read from child stream)
    UINT    nCharsHeld          = 0;        // (#of WHOLE CHARS still sitting in our buffer)
    DWORD   nBytesRead          = 0;        // (#of BYTES read from child stream)
    DWORD   nExtraBytes         = 0;        // (for properly handling half-TCHARs)
    BYTE    chExtraByte         = 0;        // (for properly handling half-TCHARs)
    BOOL    bOverflow           = FALSE;    // (TRUE == their message was longer than HOLDBUFSIZE)
    UINT    nRetcode            = 0;        // (thread function return code)

    BYTE    szReadBuff [ PIPEBUFSIZE ];     // (MUST BE EVEN! TCHARs COULD BE 2 BYTES EACH!)
    TCHAR   szHoldBuff [ HOLDBUFSIZE ];     // (MUST BE EVEN! TCHARs COULD BE 2 BYTES EACH!)

    if ( STD_ERROR_HANDLE  == nWhichStdXxx )
    {
        hOurReadFromStdXxx = pCtl->m_hOurReadFromStderr;    // stderr
        pStream  = stderr;                                  // stderr
    }
    else // ( STD_OUTPUT_HANDLE == nWhichStdXxx )
    {
        hOurReadFromStdXxx = pCtl->m_hOurReadFromStdout;    // stdout
        pStream  = stdout;                                  // stdout
    }

    const UINT nChildId = pCtl->GetChildId();  // (so we can identify which child is saying what)

    // Read child's o/p stream (pipe) and write to OUR stream instead...

    for (;;)
    {
        // Read some of the child's message data from their o/p stream (pipe)...

        szReadBuff[0] = chExtraByte;

        if ( !ReadFile
        (
            hOurReadFromStdXxx,
            &szReadBuff [ nExtraBytes ],
            PIPEBUFSIZE - nExtraBytes,
            &nBytesRead,
            NULL
        ))
        {
            if (ERROR_BROKEN_PIPE == (nRetcode = GetLastError())) nRetcode = 0;
            break;                  // (pipe forcibly closed)
        }

        if (!nBytesRead) break;     // (pipe gracefully closed)

        nBytesRead += nExtraBytes;

        nCharsRead  = nBytesRead / NUM_BYTES( TCHAR );
        nExtraBytes = nBytesRead % NUM_BYTES( TCHAR );

        if (nExtraBytes)
            chExtraByte = szReadBuff [ nBytesRead - nExtraBytes ];

        if (!nCharsRead) continue;  // (need at least ONE character!)

        // Check if message is too long for us to handle...

        if ((nCharsHeld + nCharsRead) > (HOLDBUFSIZE-1))
        {
            bOverflow = TRUE;   // (message is too long; need to truncate it)
            memcpy( &szHoldBuff[ nCharsHeld ], szReadBuff, (HOLDBUFSIZE - nCharsHeld) * NUM_BYTES( TCHAR ) );
            memcpy( szReadBuff, (LPCTSTR) g_pszBufferOverflowMsg, g_nBufferOverflowMsgLen * NUM_BYTES( TCHAR ) );
            nCharsRead = g_nBufferOverflowMsgLen;
            nCharsHeld = HOLDBUFSIZE - (g_nBufferOverflowMsgLen + 1);
        }

        // Append new data to end of hold buffer...

        memcpy( &szHoldBuff[ nCharsHeld ], szReadBuff, nCharsRead * NUM_BYTES( TCHAR ) );
        nCharsHeld += nCharsRead;
        szHoldBuff [ nCharsHeld ] = 0;  // (null terminate buffer)

        // Parse hold buffer...

        pCtl->ParseStdXxxMsg( nChildId, pStream, szHoldBuff, nCharsHeld );

        if (bOverflow) { ASSERT( !nCharsHeld ); bOverflow = FALSE; }
    }

    // WaitForProcessToExit will close our handle
    // once the process completely exits...

    return nRetcode;
}

/////////////////////////////////////////////////////////////////////////////////////////
// (MonitorStdXxxPipe helper function)

void CPipedProcessCtl::ParseStdXxxMsg( const UINT nChildId, FILE* pStream,
                                       LPTSTR pszHoldBuff, UINT& nCharsHeld )
{
    // This function executes in the context of the worker thread that calls it...

    // A worker thread that monitors a child's stdxxx o/p has received data
    // and is calling this function to parse it into newline delimited strings.

    *( pszHoldBuff + nCharsHeld ) = 0;      // (since it's string data)

    TCHAR* pbeg = pszHoldBuff;              // ptr to start of message
    TCHAR* pend = _tcschr(pbeg,_T('\n'));   // find end of message (MUST NOT BE MODIFIED!)
    if (  !pend &&
         !(pend = _tcschr(pbeg,_T('\r'))))  // find end of message (MUST NOT BE MODIFIED!)
            return;                         // we don't we have a complete message yet

    CString strMsg;                         // (work var for printing msg)
    TCHAR* pmsgend;                         // (work ptr to end of message)
                                            // (NOTE! pend variable MUST NOT BE MODIFIED!)
    do
    {
        // Remove CRLF from end of line...

        // Note: we MUST NOT MODIFY the 'pend' variable. It should always
        // point to where the newline character was found so we can start
        // looking for the next message (if there is one) where this message
        // ended.

        *pend = 0;                          // (change newline character to null)
        pmsgend = pend;                     // (start removing CRLF from here)

        while (--pmsgend >= pbeg && (*pmsgend == _T('\r') || *pmsgend == _T('\n')))
            *pmsgend = 0;

        // Write message to our stdxxx o/p stream...

        strMsg = pbeg; strMsg += _T('\n');
        m_pfnWriteToStdXXX( pStream, nChildId, strMsg );

        // 'pend' should still point to the end of this message (where newline was)

        pbeg = (pend + 1);                  // point to beg of next message (if any)
        pend = _tcschr(pbeg,_T('\n'));      // is there another message?
        if (!pend)
            pend = _tcschr(pbeg,_T('\r'));  // is there another message?
    }
    while (pend);                           // while messages remain...

    UINT nTotChars = (UINT) ((pbeg - pszHoldBuff) / NUM_BYTES( TCHAR )); // chars processed

    if (nTotChars > nCharsHeld)             // make sure we didn't process too much
    {
        ASSERT(false);                      // oops!
        nTotChars = nCharsHeld;             // fix it

        m_pfnWriteToStdXXX( stderr, nChildId, _T(PROGNAME)
            _T(": *** ParseStdXxxMsg logic error! ***\n") );
    }

    // 'Remove' the messages that we parsed from the caller's hold buffer by
    // sliding the remainder to the left (i.e. left justifying the remainder
    // in their hold buffer) and then telling them how much data now remains
    // in their hold buffer.

    // IMPORTANT PROGRAMMING NOTE! We must use memmove here and not _tcscpy!
    // _tcscpy doesn't work correctly for overlapping source and destination.

    nCharsHeld -= nTotChars;                                        // (chars that now remains)
    memmove( pszHoldBuff, pbeg, nCharsHeld * NUM_BYTES( TCHAR ) );  // (left justify remainder)
    *( pszHoldBuff + nCharsHeld ) = 0;                              // (since it's string data)
}

/////////////////////////////////////////////////////////////////////////////////////////

bool CPipedProcessCtl::WaitForProcessToExit( DWORD dwTimeOut /* =INFINITE */ )
{
    if (!m_pChildEndedThread)       // (was process ever started?)
        return true;                // (if not, then it has ended)

    if (WAIT_OBJECT_0 != WaitForSingleObject( m_pChildEndedThread->m_hThread, dwTimeOut ))
        return false;

    WaitForSingleObject( m_pStdoutThread->m_hThread, INFINITE );
    WaitForSingleObject( m_pStderrThread->m_hThread, INFINITE );

    CloseHandle( m_hChildProcess      ); m_hChildProcess      = NULL;
    CloseHandle( m_hOurReadFromStdout ); m_hOurReadFromStdout = NULL;
    CloseHandle( m_hOurReadFromStderr ); m_hOurReadFromStderr = NULL;

    return true;
}

/////////////////////////////////////////////////////////////////////////////////////////
