// Copyright (C) 2008-2012, Software Development Laboratories, "Fish" (David B. Trout)
//////////////////////////////////////////////////////////////////////////////////////////
// RunJobs.cpp:
//////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  04/19/08    1.0.0   Fish    Created.
//  11/06/08    1.0.2   Fish    Use UINT for g_nMaxRC and nExitCode instead of int since
//                              VCBUILD apparently returns "-1" on failure, which is NOT
//                              "greater than" the current nExitCode value (which starts
//                              out at zero), thereby causing RunJobs to exit with exit
//                              code 0 instead of a non-zero exit code. Thanks go out to
//                              esteban.papp for identifying this issue!
//  03/17/09    1.1.0   Fish    -rcfile support.
//  08/11/09    1.1.1   Fish    Fix CPipedProcessCtl::StartChildProcess failure return-
//                              code issue causing g_nMaxRC to not get updated.
//  08/20/09    1.1.2   Fish    Fix unique job# issue: use AssignJobNum function.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "RunJobs.h"
#include "ParseOpt.h"
#include "PipedProcessCtl.h"

/////////////////////////////////////////////////////////////////////////////////////////
// The one and only MFC application object...

CWinApp  g_App;                 // (this one object does it all; it's the entire app)
using namespace std;            // (whatever...)

/////////////////////////////////////////////////////////////////////////////////////////
// Some global variables...

CStdioFileEx  g_fRCFile;                        // (the -rcfile object itself)
bool          g_bRCFile            = false;     // (true if -rcfile specified)
UINT          g_nMaxRC             = 0;         // (maximum return code value)
HANDLE        g_hChildIdSemaphore  = NULL;      // (provides unique child-ids)

/////////////////////////////////////////////////////////////////////////////////////////
// (mainline helper function...)

void EndOfJob( CPipedProcessCtl* pJob );    // (forward reference)

/////////////////////////////////////////////////////////////////////////////////////////

int _tmain( int argc, TCHAR* argv[], TCHAR* envp[] )
{
    InitializeCriticalSection( &g_StdXxxStreamLock );

    // PROGRAMMING NOTE: In case we're running under the new VS2005 IDE, we need
    // to undefine the following variable to prevent MSBUILD/VCBUILD/CL.EXE from
    // sending UNICODE o/p directly to the IDE's o/p window instead of normal NON-
    // UNICODE o/p to stdout/stderr...

    // I don't think  there's any need to do this right away, but we might as well
    // so we don't forget and so we don't have to worry about it...

    // I'm not sure what the variable's numeric value is actually supposed to be,
    // but I'm GUESSING it's the file# or handle value of the file that's supposed
    // to be used for UNICODE o/p... <shrug> More research is obviously needed...

    _putenv (  "VS_UNICODE_OUTPUT=" );  // (use normal NON-UNICODE stdout/stderr o/p please!)
    _wputenv( L"VS_UNICODE_OUTPUT=" );  // (use normal NON-UNICODE stdout/stderr o/p please!)

    // Save the current working directory right away...  (MFC not needed)

    {
        TCHAR  szCurrDir[ 2 * MAX_PATH ];
        DWORD  dwBuffLen = NUM_TCHARS( szCurrDir );
        DWORD  dwRetCode = GetCurrentDirectory( dwBuffLen, szCurrDir );
        ASSERT( dwRetCode > 0 && dwRetCode < dwBuffLen );
        szCurrDir[ dwBuffLen - 1 ] = 0;
        g_strCurrDir = szCurrDir;
    }

    // Initialize MFC and print any error on failure...

    if (!AfxWinInit( ::GetModuleHandle( NULL ), NULL, ::GetCommandLine(), 0 ) )
    {
        _ftprintf( stderr, _T(PROGNAME) _T(": ERROR: MFC initialization FAILED!\n") );
        DeleteCriticalSection( &g_StdXxxStreamLock );
        return -1;
    }

    //-------------------------
    // Parse command-line...
    //-------------------------

    CMyCmdLineParser        parser;
    g_App.ParseCommandLine( parser );
    CString                 strMsg;

    if (!parser.m_bErrors && g_strRunJobsFilename.IsEmpty())
    {
        parser.m_bErrors = true;
        strMsg = _T(PROGNAME) _T(": ERROR: input <jobsfile> not specified or empty.\n");
        TRACE( _T("%s"), (LPCTSTR) strMsg );
        WriteToStdXxx( stderr, 0, strMsg );
    }

    if (parser.m_bErrors)
    {
        parser.ShowHelp();
        DeleteCriticalSection( &g_StdXxxStreamLock );
        return -1;
    }

    g_bRCFile = !g_strRCFileFilename.IsEmpty() ? true : false;

    //--------------------------------
    // Open input and output files...
    //--------------------------------

    CFileException  fe;     // (for catching file open errors)

    // Open the input jobs file...

    CStdioFileEx  fJobsFile;

    if (!fJobsFile.Open
    (
        g_strRunJobsFilename,
        0
            | CFile::modeRead
            | CFile::shareDenyWrite
            | CFile::typeText
            | CFile::osSequentialScan
        ,
        &fe
    ))
    {
        // Open error...

        CString strExceptionMsg;
        LPTSTR  pszExceptionMsg = strExceptionMsg.GetBuffer( 256 );
        VERIFY( fe.GetErrorMessage( pszExceptionMsg, 256 ) );
        *(pszExceptionMsg + 256 - 1) = 0;
        strExceptionMsg.ReleaseBuffer();

        strMsg.Format
        (
            _T(PROGNAME) _T(": ERROR: Could not open input jobs file \"%s\";\n\t%s.\n")

            ,g_strRunJobsFilename
            ,strExceptionMsg
        );

        TRACE( _T("%s"), (LPCTSTR) strMsg );
        WriteToStdXxx( stderr, 0, strMsg );

        DeleteCriticalSection( &g_StdXxxStreamLock );
        return -1;  // (abort)
    }

    // Open the output return-codes file...

    if (g_bRCFile)
    {
        if (!g_fRCFile.Open
        (
            g_strRCFileFilename,
            0
                | CFile::modeWrite
                | CFile::modeCreate
                | CFile::shareDenyWrite
                | CFile::typeText
            ,
            &fe
        ))
        {
            // Open error...

            CString strExceptionMsg;
            LPTSTR  pszExceptionMsg = strExceptionMsg.GetBuffer( 256 );
            VERIFY( fe.GetErrorMessage( pszExceptionMsg, 256 ) );
            *(pszExceptionMsg + 256 - 1) = 0;
            strExceptionMsg.ReleaseBuffer();

            strMsg.Format
            (
                _T(PROGNAME) _T(": ERROR: Could not open output rcfile \"%s\";\n\t%s.\n")

                ,g_strRunJobsFilename
                ,strExceptionMsg
            );

            TRACE( _T("%s"), (LPCTSTR) strMsg );
            WriteToStdXxx( stderr, 0, strMsg );

            DeleteCriticalSection( &g_StdXxxStreamLock );
            return -1;  // (abort)
        }
    }

    //-----------------------------------
    // Read jobs file and start jobs...
    //-----------------------------------

    LONG nJobNum = 0;

    CString            strJobLine;
    CString            strCmdLine;
    LPTSTR             pszCmdLine;
    DWORD              dwCmdLineLen;
    CJobsPtrList       jobsPtrList;
    CPipedProcessCtl*  pJob;
    DWORD              dwJobCount;
    HANDLE*            pJobHandles;

    while ( fJobsFile.ReadString( strJobLine ) )
    {
        // (ignore blank lines and/or comment lines...)

        strJobLine = strJobLine.Trim();

        if (0
            || strJobLine.IsEmpty()
            || strJobLine[0] == _T('#')
            || strJobLine[0] == _T(';')
            || strJobLine[0] == _T('*')
        )
            continue;

        // (count new job...)

        if ((nJobNum = AssignJobNum()) < 0)
            return -1; // (catastrophe)

        // (expand %environment% variables...)

        pszCmdLine = strCmdLine.GetBuffer( MAX_CMDLINE_LEN + 2 );

        dwCmdLineLen = ExpandEnvironmentStrings( strJobLine, pszCmdLine, MAX_CMDLINE_LEN );

        if (!dwCmdLineLen || dwCmdLineLen > MAX_CMDLINE_LEN)
        {
            strMsg.Format
            (
                _T(PROGNAME) _T(": ERROR: Could not expand cmdline for job %d = \"%s\";\n\t%s.\n")

                ,nJobNum
                ,strJobLine
                ,FormatLastError()
            );

            TRACE( _T("%s"), (LPCTSTR) strMsg );
            WriteToStdXxx( stderr, nJobNum, strMsg );

            dwCmdLineLen = dwCmdLineLen ? MAX_CMDLINE_LEN : 1;
        }

        *(pszCmdLine + dwCmdLineLen - 1) = 0;
        strCmdLine.ReleaseBuffer( dwCmdLineLen - 1 );

        //------------------------------
        // Wait for a free job slot...
        //------------------------------

        // If a limit to the #of simultaneous jobs that can be run
        // at the same time was specified, then wait until the #of
        // currently running jobs drops below the threshold before
        // starting the next one...

        dwJobCount = (DWORD) jobsPtrList.GetCount();

        while (g_dwMaxJobs && dwJobCount >= g_dwMaxJobs)
        {
#ifdef _DEBUG
            strMsg = _T(PROGNAME) _T(": INFO: Waiting for free job-slot...\n");
            TRACE( _T("%s"), (LPCTSTR) strMsg );
            WriteToStdXxx( stdout, 0, strMsg );
#endif // _DEBUG

            // Build array of job handles to wait on...

            pJobHandles = new HANDLE[ dwJobCount ];
            POSITION savepos, pos = jobsPtrList.GetHeadPosition();
            dwJobCount = 0;

            while ( pos )
            {
                pJob = jobsPtrList.GetNext( pos );
                pJobHandles[ dwJobCount ] = pJob->GetProcessHandle();
                dwJobCount++;
            }

            // Wait for any active job to finish...

            WaitForMultipleObjects( dwJobCount, pJobHandles, FALSE, INFINITE );
            delete [] pJobHandles;
#ifdef _DEBUG
            strMsg = _T(PROGNAME) _T(": INFO: a job just finished...\n");
            TRACE( _T("%s"), (LPCTSTR) strMsg );
            WriteToStdXxx( stdout, 0, strMsg );
#endif // _DEBUG

            // Free the job slot for all jobs that have just finished...

            pos = jobsPtrList.GetHeadPosition();

            while ( pos )
            {
                savepos = pos;
                pJob = jobsPtrList.GetNext( pos );

                if (pJob->WaitForProcessToExit( 10 ))
                {
                    ASSERT( pJob->HasProcessExited() );
                    EndOfJob( pJob );
                    jobsPtrList.RemoveAt( savepos );
                    pos = jobsPtrList.GetHeadPosition();
                }
            }

            dwJobCount = (DWORD) jobsPtrList.GetCount();

        } // end while (g_dwMaxJobs && dwJobCount >= g_dwMaxJobs)


        //-------------------------
        // Start this next job...
        //-------------------------

        pJob = new CPipedProcessCtl( nJobNum, strCmdLine, g_strCurrDir, &WriteToStdXxx );

        if (!pJob->StartChildProcess())
        {
            // ERROR: job could not be started! (the error message
            // should already have been issued by CPipedProcessCtl)
#ifdef _DEBUG
            strMsg.Format
            (
                _T(PROGNAME) _T(": ERROR: Job %d FAILED to start = \"%s\".\n")

                ,pJob->GetChildId()
                ,pJob->GetCommandLine()
            );
            TRACE( _T("%s"), (LPCTSTR) strMsg );
            WriteToStdXxx( stderr, 0, strMsg );
#endif // _DEBUG

            EndOfJob( pJob );  // (oh well)
        }
        else // (job successfully started)
        {
            // Add to started jobs list...
#ifdef _DEBUG
            strMsg.Format
            (
                _T(PROGNAME) _T(": INFO: Job %d started = \"%s\".\n")

                ,pJob->GetChildId()
                ,pJob->GetCommandLine()
            );
            TRACE( _T("%s"), (LPCTSTR) strMsg );
            WriteToStdXxx( stderr, 0, strMsg );
#endif // _DEBUG

            jobsPtrList.AddTail( pJob );
        }

    } // end while ( fJobsFile.ReadString( strJobLine ) )


    //-----------------------------------------------
    // Wait for all still-running jobs to finish...
    //-----------------------------------------------

    while (!jobsPtrList.IsEmpty())
    {
        // (process in REVERSE ORDER this time since jobs
        // at the end of list are the ones most likely to
        // finish last, and since we need to wait for all
        // of them to finish we might as well wait on the
        // one that's most likely to take the longest...)

        pJob = jobsPtrList.RemoveTail();

#ifdef _DEBUG
        strMsg.Format
        (
            _T(PROGNAME) _T(": INFO: Waiting for job %d to finish = \"%s\".\n")

            ,pJob->GetChildId()
            ,pJob->GetCommandLine()
        );
        TRACE( _T("%s"), (LPCTSTR) strMsg );
        WriteToStdXxx( stderr, 0, strMsg );
#endif // _DEBUG

        VERIFY( pJob->WaitForProcessToExit() );
        EndOfJob( pJob );
    }

    DeleteCriticalSection( &g_StdXxxStreamLock );

    return (int) g_nMaxRC;
}

/////////////////////////////////////////////////////////////////////////////////////////
// (mainline helper function...)

void EndOfJob( CPipedProcessCtl* pJob )
{
    // PURPOSE: retrieve the return code from the just-finished job, update our
    // max return-code value if its greater, write the return code to the return
    // codes file if needed, and then delete/discard the job object since we're
    // now completely done with it...

    // (retrieve the return code)

    UINT nExitCode = (UINT) pJob->GetProcessExitCode();

    // (save the highest one seen)

    if (nExitCode > g_nMaxRC)       // (greater than maximum?)
        g_nMaxRC = nExitCode;       // (then save new maximum)

    // Write the return code to the rcfile if specified...

    if (g_bRCFile)
    {
        CString strRCLine;
        strRCLine.Format( _T("%u = %u\n"), pJob->GetChildId(), nExitCode );
        g_fRCFile.WriteString( strRCLine );
    }

    // (discard job object now that we're done with it)

    delete pJob;        // (we don't need it any more)
}

/////////////////////////////////////////////////////////////////////////////////////////
