// Copyright (C) 2008-2012, Software Development Laboratories, "Fish" (David B. Trout)
//////////////////////////////////////////////////////////////////////////////////////////
// StdAfx.cpp: source file that includes just the standard includes
// RunJobs.pch will be the pre-compiled header
// StdAfx.obj will contain the pre-compiled type information
//////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  04/19/08    1.0.0   Fish    Created.
//  05/08/08    1.0.1   Fish    WriteToStdXxx: strip %d> prefix if nChildId supplied.
//  11/06/08    1.0.2   Fish    Use UINT instead of int.
//  03/16/09    1.1.0   Fish    WriteToStdXxx: do *NOT* strip the %d> child prefix if it
//                              is supplied, but *DO* remove the erroneous space we were
//                              inserting following our own %d> child prefix indicator.
//  03/17/09    1.1.0   Fish    -rcfile support.
//  08/20/09    1.1.2   Fish    AssignJobNum function.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"

//////////////////////////////////////////////////////////////////////////////////////////
// (return default maximum #of concurrent job processes)

DWORD MaxJobs()
{
    // Disable C4996 warning since we don't have multiple threads
    // modifying environment variables while others are trying to
    // retrieve them (which is the only reason I can think of why
    // getenv should be considered potentially unsafe)...

    #pragma warning( push )             // (save settings)
    #pragma warning( disable : 4996 )   // (disable warning C4996: "This function or variable may be unsafe...")

    LPCTSTR pszMaxJobs = _tgetenv( _T("RUNJOBS_DEFAULT_MAXJOBS") );

    #pragma warning( pop )              // (restore settings)

    if (pszMaxJobs)
    {
        int nMaxJobs = _ttoi( pszMaxJobs );
        if (nMaxJobs >= 0 && errno == 0)
            return nMaxJobs;
    }

    // (use #of system processors)

    SYSTEM_INFO     si;
    GetSystemInfo( &si );
    return si.dwNumberOfProcessors;
}

//////////////////////////////////////////////////////////////////////////////////////////

DWORD    g_dwMaxJobs    =   MaxJobs();      // (maximum #of simultaneous jobs)
CString  g_strProgName  = _T(PROGNAME);     // (our name as a CString)
CString  g_strCurrDir;                      // (current working directory)
CString  g_strRunJobsFilename;              // (input runjobs filename)
CString  g_strRCFileFilename;               // (output rcfile filename)

//////////////////////////////////////////////////////////////////////////////////////////
// (for controlling multiple threads writing to our stdout/stderr stream)

CRITICAL_SECTION  g_StdXxxStreamLock;       // (thread syncronization lock)

inline void  LockStdXxxStream()    { EnterCriticalSection( &g_StdXxxStreamLock ); };
inline void  UnlockStdXxxStream()  { LeaveCriticalSection( &g_StdXxxStreamLock ); };

//////////////////////////////////////////////////////////////////////////////////////////
// (handles asynchronous multiple threads writing to stdxxx stream at same time)

void WriteToStdXxx( FILE* pStream, UINT nChildId, LPCTSTR pszMsg )
{
    LockStdXxxStream();
    {
        if (nChildId > 0)
        {
            // PROGRAMMING NOTE: we must be careful to *NOT* insert a space
            // following the '>' character of our child-id prefix indicator
            // since doing so confuses Visual Studio error-message parsing.

            _ftprintf( pStream, _T("%d>%s"), nChildId, pszMsg );
        }
        else
        {
            _ftprintf( pStream, _T("%s"), pszMsg );
        }
        fflush( pStream );
    }
    UnlockStdXxxStream();
}

//////////////////////////////////////////////////////////////////////////////////////////
// (format last error as CString)

CString  FormatLastError ( DWORD dwLastError /* =GetLastError() */ )
{
    CString strError; LPTSTR pszMsgBuf = strError.GetBuffer( 256 );
    *pszMsgBuf = 0; DWORD dwChars = ::FormatMessage
    (
        0
            | FORMAT_MESSAGE_FROM_SYSTEM
            | FORMAT_MESSAGE_IGNORE_INSERTS
        ,
        NULL,                                       // pointer to  message source
        dwLastError,                                // requested message identifier
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),  // language identifier for requested message
        pszMsgBuf,                                  // pointer to message buffer
        256,                                        // CHARACTER count of destination buffer (NOT byte count!)
        NULL                                        // pointer to array of message inserts
    );
    *( pszMsgBuf + 256 - 1 ) = 0;
    strError.ReleaseBuffer(); if ( !dwChars ) strError = _T("");
    strError.Trim( _T("\r\n") );
    return strError;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Helper function to assign a system-wide unique job number...

LONG AssignJobNum()
{
    // TECHNIQUE: create a semaphore that will never be waited on
    // and simply use ReleaseSemaphore to increment its counter...

    // Programming note: we cannot use InterlockedIncrement here
    // since InterlockedIncrement is only process-wide (it only
    // prevents two or more threads within the same process from
    // simultaneously incrementing the same variable) whereas by
    // using a named semaphore, processes other than our own can
    // also increment it in a controlled interlocked fashion by
    // simply releasing the semaphore...

    static const LONG  nMaxJobNum   = LONG_MAX;
    static HANDLE      hSemaphore   = NULL;
    LONG               nJobNum      = 1;

    // Create/Open the semaphore if we haven't done so yet...

    if (1
        && !hSemaphore
        && !(hSemaphore = CreateSemaphore
        (
            NULL,                       // LPSECURITY_ATTRIBUTES  lpSemaphoreAttributes,
            nJobNum,                    // LONG                   lInitialCount,
            nMaxJobNum,                 // LONG                   lMaximumCount,
            _T(JOBNUM_SEMAPHORE_NAME)   // LPCTSTR                lpName
        ))
    )
    {
        DWORD dwLastError = GetLastError();
        CString strErrMsg; strErrMsg.Format
        (
            _T(PROGNAME) _T(": ERROR: CreateSemaphore() failed; rc=%u (0x%08.8X): %s\n")

            ,dwLastError,dwLastError
            ,FormatLastError( dwLastError )
        );
        TRACE( _T("%s"), (LPCTSTR) strErrMsg );
        WriteToStdXxx( stderr, 0, strErrMsg );
        ASSERT( FALSE );
        return -1;
    }

    // Release it to increment the count (and return us the previous count)...

    if (!ReleaseSemaphore( hSemaphore, +1, &nJobNum ))
    {
        // PROGRAMMING NOTE: we don't expect our maximum value to ever be
        // exceeded (even though it's technically possible the way we are
        // designed), so we purposely don't bother checking for (nor try
        // recovering from) 'ERROR_TOO_MANY_POSTS'...

        DWORD dwLastError = GetLastError();
        CString strErrMsg; strErrMsg.Format
        (
            _T(PROGNAME) _T(": ERROR: ReleaseSemaphore() failed; rc=%u (0x%08.8X): %s\n")

            ,dwLastError,dwLastError
            ,FormatLastError( dwLastError )
        );
        TRACE( _T("%s"), (LPCTSTR) strErrMsg );
        WriteToStdXxx( stderr, 0, strErrMsg );
        ASSERT( FALSE );
        return -1;
    }

    ASSERT( nJobNum > 0 );      // (sanity check)

    // Return the system-wide unique job number...

    return nJobNum;
}

//////////////////////////////////////////////////////////////////////////////////////////
