// Copyright (C) 2008-2012, Software Development Laboratories, "Fish" (David B. Trout)
/////////////////////////////////////////////////////////////////////////////////////////
//
// StdioFileEx.cpp  --  derived CStdioFile class that transparently handles UNICODE
//
/////////////////////////////////////////////////////////////////////////////////////////
//
// This class is a modified version of code originally
// written by by David Terracino <davet@lycosemail.com>:
//
// "This code is released into the public domain, because
//  it's silly easy. If you want to use and expand it, go
//  ahead; fine by me. If you do expand this class's
//  functionality, please consider sending it back up to
//  the MFC Programmer's Sourcebook at www.codeguru.com.
//
//  And as always, please give credit where credit is
//  due. Don't remove my name from the source."
//
/////////////////////////////////////////////////////////////////////////////////////////
//
// Version 1.1 23 August 2003.   Incorporated fixes from Dennis Jeryd.
// Version 1.3 19 February 2005. Incorporated fixes from Howard J Oh and some of my own.
// Version 1.4 26 February 2005. Fixed stupid screw-up in code from 1.3.
// Version 1.5 18 November 2005. - Incorporated fixes from Andy Goodwin.
//                               - Allows code page to be specified for reading/writing
//                               - Properly calculates multibyte buffer size instead of
//                                  assuming lstrlen(s).
//                               - Should handle UTF8 properly.
//
// Copyright David Pritchard 2003-2005. davidpritchard@ctv.es
//
// You can use this class freely, but please keep my ego happy
// by leaving this comment in place.
//
//////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  11/15/06    1.0.0   Fish    Removed CodePage and GetCharCount crap, remove conversion
//                              crap and use MFC "T2CA", etc, conversion macros instead,
//                              change 'modeWriteUnicode' value to not collide w/MFC.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "StdioFileEx.h"

//////////////////////////////////////////////////////////////////////////////////////////
// FIXME: little-endian presumed...

#define  UTF16LE_BOM   0xFEFF       // UNICODE file "Byte-Order Mark" (BOM)

//////////////////////////////////////////////////////////////////////////////////////////

const UINT CStdioFileEx::modeWriteUnicode = 0x40000000; // write file in UNICODE format

//////////////////////////////////////////////////////////////////////////////////////////

CStdioFileEx::CStdioFileEx(): CStdioFile()
{
    m_bIsUnicodeFile = false;
}

//////////////////////////////////////////////////////////////////////////////////////////

CStdioFileEx::CStdioFileEx( LPCTSTR pszFileName, UINT nOpenFlags )

    : CStdioFile( pszFileName, ProcessFlags( pszFileName, nOpenFlags ) )
{
}

//////////////////////////////////////////////////////////////////////////////////////////

BOOL CStdioFileEx::Open( LPCTSTR pszFileName, UINT nOpenFlags, CFileException* pError /* =NULL */ )
{
   return CStdioFile::Open( pszFileName, ProcessFlags( pszFileName, nOpenFlags ), pError );
}

//////////////////////////////////////////////////////////////////////////////////////////
// (open helper function to modify open flags if needed and set 'm_bIsUnicodeFile')

UINT CStdioFileEx::ProcessFlags( const CString& sFilePath, UINT& nOpenFlags )
{
    m_nOpenFlags = nOpenFlags;

#ifdef _DEBUG

    if ( m_nOpenFlags & CStdioFileEx::modeWriteUnicode )
    {
        ASSERT( ( m_nOpenFlags & CFile::modeWrite     ) ||
                ( m_nOpenFlags & CFile::modeReadWrite ) );
    }

#endif

    // PROGRAMMING NOTE: must not use "(nOpenFlags & modeRead)"
    //                   because 'modeRead' defined as 0x0000.

    // If reading in text mode and not creating... (fixed by Dennis Jeryd 6/8/03)

    if (1
        &&  (m_nOpenFlags & CFile::typeText)
        && !(m_nOpenFlags & CFile::modeCreate)
        && !(m_nOpenFlags & CFile::modeWrite)
    )
    {
        // Reading; we need to discover if file is UNICODE or not...

        m_bIsUnicodeFile = IsFileUnicode( sFilePath );

        // PROGRAMMING NOTE: switching to binary mode causes MFC's CStdioFile
        // 'fgets' and 'fputs' calls to NOT automatically convert text and to
        // instead write our data AS-IS, since WE WILL do the text conversion.
        // See "Unicode Stream I/O in Text and Binary Modes" in documentation.

        if (m_bIsUnicodeFile)
        {
            m_nOpenFlags ^= CFile::typeText;    // (bypasses text conversion)
            m_nOpenFlags |= CFile::typeBinary;  // (write data AS-IS instead)
        }
    }
    else
    {
        // Writing; THEY tell US whether file s/b UNICODE or not...

        m_bIsUnicodeFile = (m_nOpenFlags & modeWriteUnicode) ? true : false;
    }

    // Whatever UNICODE option may be in affect has now been saved in
    // our 'm_bIsUnicodeFile' flag, so turn off the 'modeWriteUnicode'
    // open flag if it's still on since it's OUR flag and not MFC's...

    m_nOpenFlags &= ~modeWriteUnicode;      // (since it's ours and not MFC's)
    nOpenFlags = m_nOpenFlags;              // (return modified flags)
    return m_nOpenFlags;                    // (ours/theirs identical)
}

//////////////////////////////////////////////////////////////////////////////////////////

bool CStdioFileEx::IsFileUnicode( const CString& sFilePath )
{
    CFile           file;
    CFileException  exFile;
    WCHAR           chBOM;
    bool            bIsUnicode  = false;

    if ( file.Open( sFilePath, CFile::typeBinary | CFile::modeRead, &exFile ) )
    {
        bIsUnicode = ( file.Read( &chBOM, NUM_BYTES( WCHAR )) > 0 && UTF16LE_BOM == chBOM );
        file.Close();
    }

    return bIsUnicode;
}

//////////////////////////////////////////////////////////////////////////////////////////

BOOL CStdioFileEx::ReadString( CString& rString )
{
    if (!m_bIsUnicodeFile)
        return CStdioFile::ReadString( rString );

    // Skip past BOM if at BOF...

    if (GetPosition() == 0)
    {
        WCHAR chBOM;
        Read( &chBOM, 2 );
    }

#ifdef _UNICODE  // (compiled for Unicode)

    // Defer to MFC; its UNICODE build already
    // does "_fgetts()" ==> "fgetws()" for us...

    return CStdioFile::ReadString( rString );

#else  // (NOT compiled for UNICODE)

    // Since MFC's non-UNICODE (MBCS) build always
    // does "_fgetts()" ==> "fgets()" (and we know
    // the file contains WIDE chars), we will need
    // to do the "fgetws()" for ourselves...

    #define  FGETS_CHUNKSIZE_CHARS   (256)

    BOOL    bSuccess    = FALSE;
    WCHAR*  pszStringW  = NULL;
    int     nChars      = 0;

    ULONGLONG  ulCurPos  = Seek( 0, CFile::current );

    do
    {
        rString = _T("");

        nChars += FGETS_CHUNKSIZE_CHARS;

        pszStringW  = new WCHAR [ nChars+1 ]; // (room for NULL in case complete line not read)

        // Bug fix by Dennis Jeryd 06/07/2003: initialise bSuccess

        // FIXME: fails if string if longer than MAX_LINE_CHARS

        Seek( ulCurPos, CFile::begin );

        // PROGRAMMING NOTE: we specify "nChars+1" in our 'fgetws' call
        // since it stops at specified amount-1 when a newline isn't found.

        if ((bSuccess = (fgetws( pszStringW, nChars+1, m_pStream ) ? TRUE : FALSE)))
            rString = pszStringW;

        delete [] pszStringW; pszStringW = NULL;

        if (!bSuccess && !feof(m_pStream))
        {
            // ERROR: clear the error and
            // then throw a CFileException...

            Afx_clearerr_s( m_pStream );

            AfxThrowFileException
            (
                CFileException::genericException,
                _doserrno,
                m_strFileName
            );
        }
    }
    while ( bSuccess && rString.GetLength() >= nChars && _T('\n') != rString.Right(1) );

    // ( !bSuccess || rString.GetLength() < nChars || _T('\n') == rString.Right(1) )

    // Discard end-of-line (EOL) sequence...

    int nStrLen = rString.GetLength();
    int nLenEOL = 2;

    if (0
        || (nStrLen >= (nLenEOL = 2) && _T("\r\n") == rString.Right( nLenEOL ))
        || (nStrLen >= (nLenEOL = 1) && _T("\n")   == rString.Right( nLenEOL ))
    )
        rString = rString.Left( nStrLen - nLenEOL );

    return (nStrLen ? TRUE : FALSE);

#endif  // (UNICODE)
}

//////////////////////////////////////////////////////////////////////////////////////////

void CStdioFileEx::WriteString( LPCTSTR pszString )
{
    if (!m_bIsUnicodeFile)
    {
        CStdioFile::WriteString( pszString );
        return;
    }

    // Write BOM if needed...

    if ( GetPosition() == 0 )
    {
        // FIXME: little-endian presumed.

        static const WCHAR chBOM = UTF16LE_BOM;
        CFile::Write( &chBOM, 2 );
    }

    // Now write out the converted string AS-IS...

    CStringW  strStringW  = pszString;      // (convert to WIDE)

    CFile::Write( (LPCWSTR)strStringW, (UINT) strStringW.GetLength() * NUM_BYTES( WCHAR ) );
}

//////////////////////////////////////////////////////////////////////////////////////////
